from concurrent.futures import ALL_COMPLETED, ProcessPoolExecutor, ThreadPoolExecutor, wait
from multiprocessing import cpu_count, freeze_support
from time import sleep, perf_counter
import time
from math import ceil, log
from typing import Dict, Type, Union
import numpy as np

from time import sleep
from unittest import TestCase, TestSuite, TextTestRunner

try:
    from KnkPoolExecutor import PipeProcessPoolExecutor, StackTracedProcessPoolExecutor, StackTracedThreadPoolExecutor
except:
    import sys
    import os
    # Add src dir to import path for debugging
    sys.path.insert(0, os.path.abspath(
        os.path.join(os.path.dirname(__file__), '..', 'src')))
    print('Unable to import, use src instead', sys.path)

    from KnkPoolExecutor import PipeProcessPoolExecutor, StackTracedProcessPoolExecutor, StackTracedThreadPoolExecutor


class CalibrateTime:
    t_sleep: float = 1
    t_np: float = 0
    t_gil: float = 0
    LOOP = 100000000

    def __init__(self, loop: int = 100000000) -> None:
        self.LOOP = loop
        print('sleep')
        t1 = perf_counter()
        sleep(1)
        t2 = perf_counter()
        self.t_sleep = t2-t1

        print('NumPy')
        t1 = perf_counter()
        self.np()
        t2 = perf_counter()
        self.t_np = t2-t1

        print('GIL math')
        t1 = perf_counter()
        self.gil()
        t2 = perf_counter()
        self.t_gil = t2-t1
        print('sleep: 1', '\tused time:', self.t_sleep,
              'np:', self.t_np, '\tGIL:', self.t_gil)

    @classmethod
    def np(cls, *a):
        return np.sum(np.log(1+np.arange(cls.LOOP, dtype=float)))

    @classmethod
    def gil(cls, *a):
        return sum(log(i+1) for i in range(cls.LOOP))


class TestPoolTime(TestCase):
    def __init__(self, methodName: str, pool: Union[Type[ThreadPoolExecutor], Type[ProcessPoolExecutor]], pool_worker: int = cpu_count(), test_loop: int = cpu_count(), calibrate_time: CalibrateTime = None, overhead_time: float = 1) -> None:
        freeze_support()
        self.pool = None
        self.pool_class = pool
        self.worker = pool_worker

        self.test_loop = test_loop
        self.test_round = ceil(test_loop/pool_worker)
        self.calibrate_time = CalibrateTime() if calibrate_time is None else calibrate_time
        self.overhead = overhead_time
        super().__init__(methodName)

    def setUp(self) -> None:
        self.pool = self.pool_class(self.worker)
        return super().setUp()

    def tearDown(self) -> None:
        if self.pool is not None:
            self.pool.shutdown(wait=True)
            self.pool = None
        return super().tearDown()

    def print_result(self, t: float, calibrate: float, func: str = None) -> float:
        if func is None:
            func = self._testMethodName
        print('*'*20)
        print('Func:', func, '\tPool:', self.pool_class.__name__, '\tWorker:',
              self.worker, '\tTestLoop:', self.test_loop, '\tWorkerRound:', self.test_round, '\tOverhead:', self.overhead)
        print('Calibrate time:', calibrate,
              'Calibrate * Round:', calibrate * self.test_round)
        print('Used time:', t)
        print('Per worker:', t/self.worker, '\tPer worker - set overhead:',
              (t-self.overhead)/self.worker, '\tEst. overhead:', t-(calibrate*self.worker))
        print('Per worker round:', t/self.test_round, '\tPer worker round - set overhead:',
              (t-self.overhead)/self.test_round, '\tEst. overhead:', t-(calibrate*self.test_round))
        print('Per jobs:', t/self.test_loop, '\tPer jobs - set overhead:',
              (t-self.overhead)/self.test_loop, '\tEst. overhead:', t-(calibrate*self.test_loop))

        self.assertGreaterEqual(
            t, calibrate, 'Jobs duration is shorter than duration of single jobs in main thread.')
        t = (t-self.overhead)/calibrate
        self.assertLessEqual(
            t, self.test_loop, 'Jobs duration is longer than running in main thread.')
        return (t-self.overhead)/self.test_loop

    def test_sleep(self):
        arg = [1] * self.test_loop
        t1 = perf_counter()
        list(self.pool.map(sleep, arg))
        t2 = perf_counter()

        self.assertGreaterEqual(t2-t1, 1, 'Sleep is shorter than 1.')

        self.assertLessEqual(
            self.print_result(t2-t1, self.calibrate_time.t_sleep), self.test_round, 'Jobs (sleep) not done in parallel ({}).'.format(self.pool_class.__name__))

    def test_np(self):
        arg = [None] * self.test_loop
        t1 = perf_counter()
        list(self.pool.map(CalibrateTime.np, arg))
        t2 = perf_counter()

        self.assertLessEqual(
            self.print_result(t2-t1, self.calibrate_time.t_np), self.test_round, 'Jobs (NumPy) not done in parallel ({}).'.format(self.pool_class.__name__))

    def test_gil(self):
        arg = [None] * self.test_loop
        t1 = perf_counter()
        list(self.pool.map(CalibrateTime.gil, arg))
        t2 = perf_counter()

        if not isinstance(self.pool, ThreadPoolExecutor):
            self.assertLessEqual(
                self.print_result(t2-t1, self.calibrate_time.t_gil), self.test_round, 'Jobs (GIL math) not done in parallel ({}).'.format(self.pool_class.__name__))


if __name__ == '__main__':
    freeze_support()

    print('Baseline')
    calibrate = CalibrateTime()

    cpu = cpu_count()
    suite = TestSuite()
    for t in ['test_sleep', 'test_np', 'test_gil']:
        for m, a in [[0, 1], [0, cpu//2], [1, 0], [1, 1], [1, cpu//2], [2, 0]]:
            for p in [ProcessPoolExecutor, ThreadPoolExecutor, StackTracedProcessPoolExecutor, StackTracedThreadPoolExecutor, PipeProcessPoolExecutor]:
                suite.addTest(
                    TestPoolTime(methodName=t, pool=p, test_loop=(
                        cpu*m)+a, calibrate_time=calibrate))

    TextTestRunner().run(suite)
    '''with StackTracedThreadPoolExecutor() as pool:
        t1str = perf_counter()
        r = [pool.submit(sleep, 1) for _ in range(LOOP)]
        wait(r, return_when=ALL_COMPLETED)
        for i in r:
            print(i.result())
        t1stp = perf_counter()
        print("In StackTracedThreadPoolExecutor sleep : --- %s seconds ---" %
              (t1stp - t1str))

        t2str = perf_counter()
        wnp = [pool.submit(np.log, i) for i in range(1, 1+LOOP)]
        wait(wnp, return_when=ALL_COMPLETED)
        for i in wnp:
            print(i.result())
        t2stp = perf_counter()
        print("In StackTracedThreadPoolExecutor np : --- %s seconds ---" %
              (t2stp - t2str))

        t3str = perf_counter()
        gil = [pool.submit(log, i) for i in range(1, 1+LOOP)]
        wait(gil, return_when=ALL_COMPLETED)
        for i in gil:
            print(i.result())
        t3stp = perf_counter()
        print("In StackTracedThreadPoolExecutor gil : --- %s seconds ---" %
              (t3stp - t3str))

    with StackTracedProcessPoolExecutor() as pool:
        r = [pool.submit(sleep, 1) for _ in range(LOOP)]
        wait(r, return_when=ALL_COMPLETED)
        t1str = perf_counter()
        for i in r:
            print(i.result())
        t1stp = perf_counter()
        print("In StackTracedProcessPoolExecutor sleep : --- %s seconds ---" %
              (t1stp - t1str))

        wnp = [pool.submit(np.log, i) for i in range(1, 1+LOOP)]
        wait(wnp, return_when=ALL_COMPLETED)
        t2str = perf_counter()
        for i in wnp:
            print(i.result())
        t2stp = perf_counter()
        print("In StackTracedProcessPoolExecutor np : --- %s seconds ---" %
              (t2stp - t2str))

        gil = [pool.submit(log, i) for i in range(1, 1+LOOP)]
        wait(gil, return_when=ALL_COMPLETED)
        t3str = perf_counter()
        for i in gil:
            print(i.result())
        t3stp = perf_counter()
        print("In StackTracedProcessPoolExecutor gil : --- %s seconds ---" %
              (t3stp - t3str))

    with PipeProcessPoolExecutor() as pool:
        r = [pool.submit(sleep, 1) for _ in range(LOOP)]
        wait(r, return_when=ALL_COMPLETED)
        t1str = perf_counter()
        for i in r:
            print(i.result())
        t1stp = perf_counter()
        print("In PipeProcessPoolExecutor sleep : --- %s seconds ---" %
              (t1stp - t1str))

        wnp = [pool.submit(np.log, i) for i in range(1, 1+LOOP)]
        wait(wnp, return_when=ALL_COMPLETED)
        t2str = perf_counter()
        for i in wnp:
            print(i.result())
        t2stp = perf_counter()
        print("In PipeProcessPoolExecutor np : --- %s seconds ---" %
              (t2stp - t2str))

        gil = [pool.submit(log, i) for i in range(1, 1+LOOP)]
        wait(gil, return_when=ALL_COMPLETED)
        t3str = perf_counter()
        for i in gil:
            print(i.result())
        t3stp = perf_counter()
        print("In PipeProcessPoolExecutor gil : --- %s seconds ---" %
              (t3stp - t3str))'''



__author__ = 'Tanapat Kahabodeekanokkul (GitHub @pahntanapat, pahntanapat@gmail.com), Natthaphong Ratchakhom (GitHub @natthaphong17)'

from concurrent.futures import (
    CancelledError,
    Executor,
    ThreadPoolExecutor,
    ProcessPoolExecutor,
    Future,
)
from concurrent.futures._base import (
    FINISHED,
    CANCELLED,
    CANCELLED_AND_NOTIFIED,
)
from concurrent.futures.thread import (
    _WorkItem,
    BrokenThreadPool,
    _shutdown,
)

import gc
from multiprocessing import Pipe, freeze_support, get_context, cpu_count
from multiprocessing.connection import Connection
from queue import Queue
import sys
import time
import traceback
from typing import Any, Callable, Iterable, Iterator, Optional, Tuple, TypeVar, Union

gc.enable()

_T = TypeVar("_T")


def function_wrapper(fn, *args, **kwargs):
    """Wraps `fn` in order to preserve the traceback of any kind of
    raised exception

    """
    try:
        return fn(*args, **kwargs)
    except Exception:
        raise sys.exc_info()[0](traceback.format_exc())  # Creates an
        # exception of the
        # same type with the
        # traceback as
        # message


class StackTraceExecutor(Executor):

    """Base class for Executor, attaching stack trace when execption occurs.

    We acknowledge [se7entyse7en](https://stackoverflow.com/users/3276106/se7entyse7en)'s [answer in StackOverflow](https://stackoverflow.com/a/24457608) about Stack Traces and [bounded-pool-executor](https://pypi.org/project/bounded-pool-executor/) about maximal limit of work queue.
    """

    def submit(self, fn, *args, **kwargs):
        """Submits the wrapped function instead of `fn`"""

        return super().submit(function_wrapper, fn, *args, **kwargs)

    def map_args(self,
                 fn: Callable[..., _T],
                 iterable: Iterable[Any],
                 *args,
                 timeout: Optional[float] = None,
                 chunksize: Optional[int] = None) -> Iterator[_T]:
        l = len(iterable)
        iterables = [iterable] + [[i] * l for i in args]
        return super().map(fn,
                           *iterables,
                           timeout=timeout,
                           chunksize=chunksize)

    def map(self, fn: Callable[..., _T], *iterables, timeout: Optional[float] = None, chunksize: int = 1) -> Iterator[_T]:
        """Returns an iterator equivalent to map(fn, iter).

        Args:
            fn: A callable that will take as many arguments as there are
                passed iterables.
            timeout: The maximum number of seconds to wait. If None, then there
                is no limit on the wait time.
            chunksize: The size of the chunks the iterable will be broken into
                before being passed to a child process. This argument is only
                used by ProcessPoolExecutor; it is ignored by
                ThreadPoolExecutor.

        Returns:
            An iterator equivalent to: map(func, *iterables) but the calls may
            be evaluated out-of-order.

        Raises:
            TimeoutError: If the entire result iterator could not be generated
                before the given timeout.
            Exception: If fn(*args) raises for any values.
        """
        if timeout is not None:
            end_time = timeout + time.monotonic()

        fs = [self.submit(fn, *args) for args in zip(*iterables)]

        # Yield must be hidden in closure so that the futures are submitted
        # before the first iterator value is required.
        def result_iterator():
            try:
                # reverse to keep finishing order
                fs.reverse()
                while fs:
                    # Careful not to keep a reference to the popped future
                    if timeout is None:
                        yield fs.pop().result()
                    else:
                        yield fs.pop().result(end_time - time.monotonic())
            except Exception:
                raise sys.exc_info()[0](traceback.format_exc())
            finally:
                for future in fs:
                    future.cancel()
        return result_iterator()


class StackTracedThreadPoolExecutor(StackTraceExecutor, ThreadPoolExecutor):
    def __init__(self, max_workers: Optional[int] = None, thread_name_prefix: str = '',
                 initializer: Callable[..., None] = None, initargs: Tuple = (), max_work_queue_size: int = 0):
        """Initializes a new ThreadPoolExecutor instance.

        Args:
            max_workers: The maximum number of threads that can be used to
                execute the given calls.
            thread_name_prefix: An optional name prefix to give our threads.
            initializer: A callable used to initialize worker threads.
            initargs: A tuple of arguments to pass to the initializer.
        """
        super().__init__(max_workers=max_workers, thread_name_prefix=thread_name_prefix,
                         initializer=initializer, initargs=initargs)
        self._work_queue = Queue(maxsize=max_work_queue_size)
        gc.collect()


class StackTracedProcessPoolExecutor(StackTraceExecutor, ProcessPoolExecutor):
    def __init__(self, max_workers: Optional[int] = None,
                 initializer: Callable[..., None] = None, initargs: Tuple = (), max_work_queue_size: int = 0):
        """Initializes a new ThreadPoolExecutor instance.

        Args:
            max_workers: The maximum number of threads that can be used to
                execute the given calls.
            thread_name_prefix: An optional name prefix to give our threads.
            initializer: A callable used to initialize worker threads.
            initargs: A tuple of arguments to pass to the initializer.
        """
        super().__init__(max_workers=max_workers,
                         initializer=initializer, initargs=initargs)
        self._work_ids = Queue(maxsize=max_work_queue_size)
        gc.collect()


class PipePoolFuture(Future):
    def cancel(self):
        """Cancel the future if possible.

        Returns True if the future was cancelled, False otherwise. A future
        cannot be cancelled if it is running or has already completed.
        """
        with self._condition:
            if self._state in {FINISHED}:
                return False

            if self._state in {CANCELLED, CANCELLED_AND_NOTIFIED}:
                return True

            self._state = CANCELLED
            self._condition.notify_all()

        self._invoke_callbacks()
        return True


class _ProcessWorkItem(_WorkItem):
    THREAD_SLEEP = 0.001

    def __init__(
        self,
        future: PipePoolFuture,
        fn: Callable,
        # logger: Logger,
        args,
        kwargs,
        initializer: Optional[Callable] = None,
        initargs: Tuple[Any, ...] = tuple(),
        mp_method: Optional[str] = None,
        thread_sleep: Optional[float] = 0.001
    ):
        self.future = future
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self.kwargs['initializer'] = initializer
        self.kwargs['initargs'] = initargs
        #self.logger = logger
        self.mp_method = mp_method

        if (thread_sleep) and (thread_sleep > 0):
            self.THREAD_SLEEP = thread_sleep

    def run(self):
        if not self.future.set_running_or_notify_cancel():
            return

        try:
            main, sub = Pipe()
            proc = get_context(self.mp_method).Process(
                target=self.process_worker,
                args=[self.fn, sub] + list(self.args),
                kwargs=self.kwargs,
            )
            proc.start()

            while proc.is_alive() and not (main.poll(None) or self.future.cancelled()):
                proc.join(self.THREAD_SLEEP)

            if main.poll(0):
                status, result = main.recv()

                if status:
                    self.future.set_result(result)
                else:
                    self.future.set_exception(result)

                main.close()
            elif self.future.cancelled():
                if proc.is_alive():
                    proc.terminate()
                    status = "The {mp} process of {fn} (pid: {pid}) ended by future.cancel() with exitcode: {exitcode}."
                    kws = dict(mp=self.mp_method,
                               fn=self.fn.__name__,
                               pid=proc.pid, exitcode=proc.exitcode())

                    #self.logger.info(status, **kws)
                    raise CancelledError(status.format_map(kws))
                raise CancelledError()
            else:
                raise RuntimeError(
                    "The {mp} process of {fn} (pid: {pid}) ended with code {code} by unknown reason - {fn}(*{arg},**{kw}).".format(
                        mp=self.mp_method,
                        fn=self.fn.__name__,
                        arg=self.args,
                        kw=self.kwargs,
                        pid=proc.pid,
                        code=proc.exitcode(),
                    )
                )

        except:
            # self.logger.exception('Exception from run process: {}', exc)
            self.future.set_exception(
                sys.exc_info()[0](traceback.format_exc()))
            # Break a reference cycle with the exception 'exc'
            self = None

    @staticmethod
    def process_worker(fn: Callable, pipe: Connection, /, *arg, initializer: Optional[Callable] = None, initargs: Tuple[Any, ...] = tuple(), **kw):
        try:
            if initializer is not None:
                initializer(*initargs)
            result = fn(*arg, **kw)
            pipe.send([True, result])
        except:
            pipe.send([False, sys.exc_info()[0](traceback.format_exc())])

        pipe.close()
        return gc.collect()


class PipeProcessPoolExecutor(StackTracedThreadPoolExecutor):
    """ProcessPoolExecutor for environment without /dev/shm
    PipeProcessPoolExecutor enable multi-cpu parallel processing in Python as same as ProcessPoolExecutor in environment without /dev/shm (shared memory for processes) support, i.e. AWS Lambda.

    Acknowlegdement
    ---------------
     - [Parallel Processing in Python with AWS Lambda, AWS Compute Blog](https://aws.amazon.com/th/blogs/compute/parallel-processing-in-python-with-aws-lambda/)
     - [AWS Lambda Memory Vs CPU configuration](https://stackoverflow.com/a/66523153)


    """

    def __init__(
        self,
        max_workers: Union[int, None] = None,
        thread_name_prefix: Optional[str] = None,
        initializer: Optional[Callable[..., None]] = None,
        initargs: Optional[Tuple[Any, ...]] = None,
        # logger: Optional[Logger] = None,
        mp_method: Optional[str] = None,
        thread_sleep: Optional[float] = None,
        max_work_queue_size: int = 0
    ) -> None:

        if max_workers is None:
            max_workers = 1 + (cpu_count() * 2)

        super().__init__(max_workers=max_workers, thread_name_prefix=thread_name_prefix,
                         initializer=initializer, initargs=initargs, max_work_queue_size=max_work_queue_size)
        # self.logger = getLogger() if (logger is None) else logger
        self.mp_method = mp_method
        self.thread_sleep = thread_sleep

    def submit(self, fn: Callable, *args, **kwargs) -> Future:
        with self._shutdown_lock:
            if self._broken:
                raise BrokenThreadPool(self._broken)

            if self._shutdown:
                raise RuntimeError(
                    "cannot schedule new futures after shutdown")
            if _shutdown:
                raise RuntimeError(
                    "cannot schedule new futures after interpreter shutdown"
                )

            f = PipePoolFuture()
            w = _ProcessWorkItem(f, fn,
                                 # self.logger,
                                 args, kwargs,
                                 initializer=self._initializer,
                                 initargs=self._initargs,
                                 mp_method=self.mp_method,
                                 thread_sleep=self.thread_sleep)

            self._work_queue.put(w)
            self._adjust_thread_count()
            return f


class Pool:
    def __init__(
        self,
        process_pool: Union[None, bool, ThreadPoolExecutor,
                            ProcessPoolExecutor] = None,
        pool_worker: Optional[int] = None,
        aws_lambda: Optional[bool] = None,
    ) -> None:

        freeze_support()
        cpu = cpu_count()
        if process_pool is None:
            if pool_worker == 0:
                self.pool = None
                return
            process_pool = cpu > 1

        if isinstance(process_pool, bool):
            if process_pool:
                if (pool_worker is None) or (pool_worker == 0):
                    pool_worker = cpu
                elif pool_worker < 0:
                    pool_worker = max(1, pool_worker - cpu_count())

                try:
                    if aws_lambda:
                        self.pool = PipeProcessPoolExecutor(
                            max_worker=pool_worker)
                    else:
                        self.pool = StackTracedProcessPoolExecutor(
                            max_workers=pool_worker)
                except OSError:
                    self.pool = PipeProcessPoolExecutor(
                        max_worker=pool_worker)
            else:
                if (pool_worker is None) or (pool_worker == 0):
                    pool_worker = min(1 + (cpu * 4), 61)
                elif pool_worker < 0:
                    pool_worker = min(max(1, (4 * cpu) + 1 - pool_worker), 61)

                self.pool = StackTracedThreadPoolExecutor(
                    max_workers=pool_worker)

        else:
            self.pool = process_pool

    def __del__(self):
        if self.pool is None:
            self.pool.shutdown(wait=False)

    def submit(self, fn: Callable, *arg, **kw) -> Future:
        return self.pool.submit(fn, *arg, **kw)



from time import perf_counter
from traceback import print_exception


class CacheProcessTime:
    start: float = None
    stop: float = None
    duration: float = None

    def __enter__(self):
        self.stop = None
        self.duration = None
        self.start = perf_counter()

    def __exit__(self, exc_type, exc_value, tb):
        self.stop = perf_counter()
        self.duration = self.stop - self.start
        if exc_type is not None:
            print_exception(exc_type, exc_value, tb)



try:
    from KnkPoolExecutor import PipeProcessPoolExecutor, StackTracedProcessPoolExecutor, StackTracedThreadPoolExecutor
except:
    import sys
    import os
    # Add src dir to import path for debugging
    sys.path.insert(0, os.path.abspath(
        os.path.join(os.path.dirname(__file__), '..', 'src')))
    print('Unable to import, use src instead', sys.path)

    from KnkPoolExecutor import PipeProcessPoolExecutor, StackTracedProcessPoolExecutor, StackTracedThreadPoolExecutor

from multiprocessing import cpu_count, freeze_support
from typing import Type, Union

from concurrent.futures import ALL_COMPLETED, ProcessPoolExecutor, ThreadPoolExecutor, wait
from time import sleep
from unittest import TestCase, TestSuite, TextTestRunner


class TestPoolResult(TestCase):
    def __init__(self, methodName: str, pool: Union[Type[ThreadPoolExecutor], Type[ProcessPoolExecutor]], pool_size: int = cpu_count()) -> None:
        freeze_support()
        self.pool = None
        self.pool_class = pool
        self.pool_size = pool_size
        super().__init__(methodName)

    def setUp(self) -> None:
        self.pool = self.pool_class(self.pool_size)
        return super().setUp()

    def tearDown(self) -> None:
        if self.pool is not None:
            self.pool.shutdown(wait=True)
            self.pool = None
        return super().tearDown()

    def test_map_sleep(self):
        for k, v in enumerate(self.pool.map(sleep, [1]*self.pool_size)):
            self.assertIsNone(
                v, 'pool.map(sleep, [1]*pool_size) at index: '+str(k))

    def test_submit_sleep(self):
        ft = [self.pool.submit(sleep, 1) for _ in range(self.pool_size)]
        wait(ft, return_when=ALL_COMPLETED)
        for k, v in enumerate(ft):
            self.assertIsNone(
                v.exception(), 'self.pool.submit(sleep, 1) raises exception at index: '+str(k))
            if v.exception() is None:
                self.assertIsNone(
                    v.result(), 'pool.map(sleep, [1]*pool_size) at index: '+str(k))


if __name__ == '__main__':
    suite = TestSuite()
    for i in {'test_map_sleep', 'test_submit_sleep'}:
        for p in {ProcessPoolExecutor, ThreadPoolExecutor, StackTracedProcessPoolExecutor, StackTracedThreadPoolExecutor, PipeProcessPoolExecutor}:
            suite.addTest(TestPoolResult(i, p))
    TextTestRunner().run(suite)
